import json
import os
import tempfile
from datetime import datetime

import configargparse

from pynYNAB.Client import clientfromargs
from pynYNAB.Entity import ComplexEncoder
from pynYNAB.schema.budget import Transaction
from pynYNAB.scripts.ofximport import do_ofximport
from test_live.common_Live import commonLive


class TestOFX(commonLive):
    def testiimport(self):
        parser = configargparse.getArgumentParser('pynYNAB')
        args=parser.parse_known_args()[0]
        args.ofxfile = os.path.join(tempfile.gettempdir(),'data.ofx')
        args.logginglevel= 'debug'

        self.client=clientfromargs(args)

        content="""OFXHEADER:100
DATA:OFXSGML
VERSION:102
SECURITY:NONE
ENCODING:USASCII
CHARSET:1252
COMPRESSION:NONE
OLDFILEUID:NONE
NEWFILEUID:NONE
<OFX>
<SIGNONMSGSRSV1>
<SONRS>
<STATUS>
<CODE>0
<SEVERITY>INFO
</STATUS>
<DTSERVER>20130313133728
<LANGUAGE>FRA
</SONRS>
</SIGNONMSGSRSV1>
<BANKMSGSRSV1>
<STMTTRNRS>
<TRNUID>29939615002
<STATUS>
<CODE>0
<SEVERITY>INFO
</STATUS>
<STMTRS>
<CURDEF>EUR
<BANKACCTFROM>
<BANKID>11706
<BRANCHID>41029
<ACCTID>29939615002
<ACCTTYPE>CHECKING
</BANKACCTFROM>
<BANKTRANLIST>
<DTSTART>20130128000000
<DTEND>20130314235959
<STMTTRN>
<TRNTYPE>CHECK
<DTPOSTED>20130312
<TRNAMT>-491.09
<FITID>13071099780237330004
<CHECKNUM>0003445
<NAME>CHEQUE
<MEMO>CHEQUE
</STMTTRN>
</BANKTRANLIST>
<LEDGERBAL>
<BALAMT>-6348.01
<DTASOF>20130312
</LEDGERBAL>
<AVAILBAL>
<BALAMT>-6348.01
<DTASOF>20130312
</AVAILBAL>
</STMTRS>
</STMTTRNRS>
</BANKMSGSRSV1>
</OFX>"""
        with open(args.ofxfile,'w') as f:
            f.writelines(content)
        imported_date=datetime.now().date()

        testaccount='TEST'
        account=self.util_get_empty_account_by_name_if_doesnt_exist(testaccount)

        key = '11706 41029 29939615002'
        account.note='great note key[%s]key' %key
        self.client.budget.be_accounts.modify(account)
        self.client.sync()

        def getTr(date,payee,amount,memo,account):
            return Transaction(
                entities_account_id=account.id,
                date=date,
                entities_payee_id=self.util_add_payee_by_name_if_doesnt_exist(payee).id,
                imported_payee=payee,
                source='Imported',
                check_number='0003445',
                memo=memo,
                amount=amount,
                cash_amount=amount,
                imported_date=imported_date
            )
        Transactions = [
            getTr(datetime(year=2013, month=3, day=12).date(),'CHEQUE',-491.09,'CHEQUE    13071099780237330004',account),
        ]

        do_ofximport(args)
        self.reload()
        for tr in Transactions:
            print('Should have been imported:')
            print(json.dumps(tr,cls=ComplexEncoder))
            print('Found in the register:')
            print(json.dumps([tr2 for tr2 in self.client.budget.be_transactions if tr2.amount==tr.amount],cls=ComplexEncoder))
            self.assertTrue(self.client.budget.be_transactions.containsduplicate(tr),
                            msg='couldnt find a transaction with the same hash after ofx import')



import configargparse

from pynYNAB.Client import clientfromargs
from pynYNAB.Entity import UnknowEntityFieldValueError

parser = configargparse.getArgumentParser('pynYNAB')
args = parser.parse_known_args()[0]

try:
    client = clientfromargs(args)
    client.sync()
except UnknowEntityFieldValueError,e:
    print(e.message)



import codecs
import inspect
import json
from collections import namedtuple
from datetime import datetime
import configargparse
import os
import csv
import sys

import jsontableschema
from jsontableschema.exceptions import InvalidSchemaError
from jsontableschema.model import SchemaModel

from pynYNAB.Client import clientfromargs
from pynYNAB.schema.budget import Payee, Transaction
from pynYNAB.scripts.config import get_logger, test_common_args

scriptsdir=os.path.dirname(os.path.abspath(__file__))
schemas_dir = os.path.join(scriptsdir,'csv_schemas')

def csvimport_main():
    print('pynYNAB CSV import')
    """Manually import a CSV into a nYNAB budget"""
    parser = configargparse.getArgumentParser('pynYNAB')
    parser.description=inspect.getdoc(csvimport_main)
    parser.add_argument('csvfile', metavar='CSVpath', type=str,
                        help='The CSV file to import')
    parser.add_argument('schema', metavar='schemaName', type=str,
                        help='The CSV schema to use (see csv_schemas directory)')
    parser.add_argument('accountname', metavar='AccountName', type=str,nargs='?',
                        help='The nYNAB account name  to use')
    parser.add_argument('-import-duplicates', action='store_true',
                        help='Forces the import even if a duplicate (same date, account, amount, memo, payee) is found')

    args = parser.parse_args()
    test_common_args(args)

    if not os.path.exists(args.csvfile):
        get_logger().error('input CSV file does not exist')
        exit(-1)

    do_csvimport(args)


def do_csvimport(args,client=None):
    if client is None:
        client = clientfromargs(args)
    logger=get_logger(args)

    logger.debug('selected schema %s' % (args.schema,))
    if os.path.exists(args.schema):
        schemafile = args.schema
    else:
        schemafile = os.path.join(schemas_dir, args.schema + '.json')
        if not os.path.exists(schemafile):
            logger.error('This schema doesn''t exist in csv_schemas')
            exit(-1)
    try:
        schema = SchemaModel(schemafile, case_insensitive_headers=True)
        with open(schemafile,'r') as sf:
            schemacontent = json.load(sf)
            try:
                nheaders = schemacontent['nheaders']
            except KeyError:
                nheaders = 1
    except InvalidSchemaError:
        logger.error('Invalid CSV schema')
        raise
    logger.debug('schema headers %s' % schema.headers)

    if 'account' not in schema.headers and args.accountname is None:
        logger.error('This schema does not have an account column and no account name was provided')
        exit(-1)

    accounts = {x.account_name: x for x in client.budget.be_accounts}
    payees = {p.name: p for p in client.budget.be_payees}
    mastercategories_perid = {m.id: m for m in client.budget.be_master_categories}
    subcategories = {}
    for s in client.budget.be_subcategories:
        m=mastercategories_perid[s.entities_master_category_id]
        subcategories[m.name+':'+s.name]=s

    def getaccount(accountname):
        try:
            logger.debug('searching for account %s' % accountname)
            return accounts[accountname]
        except KeyError:
            logger.error('Couldn''t find this account: %s' % accountname)
            exit(-1)

    def getpayee(payeename):
        try:
            logger.debug('searching for payee %s' % payeename)
            return payees[payeename]
        except KeyError:
            logger.debug('Couldn''t find this payee: %s' % payeename)
            payee=Payee(name=payeename)
            client.budget.be_payees.append(payee)
            return payee

    def getsubcategory(categoryname):
        try:
            logger.debug('searching for subcategory %s' % categoryname)
            return subcategories[categoryname]
        except KeyError:
            get_logger(args).debug('Couldn''t find this category: %s' % categoryname)
            exit(-1)

    if 'account' not in schema.headers:
        entities_account_id = getaccount(args.accountname).id

    if 'inflow' in schema.headers and 'outflow' in schema.headers:
        pass
    elif 'amount' in schema.headers:
        pass
    else:
        logger.error('This schema doesn''t provide an amount column or (inflow,outflow) columns')
        exit(-1)

    csvrow = namedtuple('CSVrow', field_names=schema.headers)
    transactions = []

    imported_date=datetime.now().date()

    get_logger(args).debug('OK starting the import from %s '%os.path.abspath(args.csvfile))
    with open(args.csvfile, 'r') as inputfile:
        header = inputfile.readline()
        for row in csv.reader(inputfile):
            if sys.version[0] == '2':
                row = [cell.decode('utf-8') for cell in row]
            if all(map(lambda x:x.strip()=='',row)):
                continue
            get_logger(args).debug('read line %s' % row)
            result = csvrow(*list(schema.convert_row(*row, fail_fast=True)))
            if 'account' in schema.headers:
                entities_account_id = getaccount(result.account).id
            if 'inflow' in schema.headers and 'outflow' in schema.headers:
                amount = result.inflow - result.outflow
            elif 'amount' in schema.headers:
                amount = result.amount
            else:
                get_logger(args).error('Couldn''t find this account: %s' % args.accountname)
                exit(-1)

            if 'category' in schema.headers and result.category:
                entities_subcategory_id = getsubcategory(result.category).id
            else:
                entities_subcategory_id = None
            if 'payee' in schema.headers:
                imported_payee=result.payee
            else:
                imported_payee=''
            entities_payee_id = getpayee(imported_payee).id
            if 'memo' in schema.headers:
                memo=result.memo
            else:
                memo=''


            transaction=Transaction(
                entities_account_id=entities_account_id,
                amount=amount,
                date=result.date,
                entities_payee_id=entities_payee_id,
                entities_subcategory_id=entities_subcategory_id,
                imported_date=imported_date,
                imported_payee=imported_payee,
                memo=memo,
                source="Imported"
            )
            if args.import_duplicates or (not client.budget.be_transactions.containsduplicate(transaction)):
                get_logger(args).debug('Appending transaction %s '%transaction.getdict())
                transactions.append(transaction)
            else:
                get_logger(args).debug('Duplicate transaction found %s '%transaction.getdict())



    client.add_transactions(transactions)

if __name__ == "__main__":
    csvimport_main()



import copy
import json
import logging

from aenum import Enum

from pynYNAB import KeyGenerator
from pynYNAB.schema.Fields import EntityField, EntityListField, PropertyField

logger = logging.getLogger('pynYNAB')

def undef():
    pass


class AccountTypes(Enum):
    undef = 'undef'
    Checking = 'Checking'
    Savings = 'Savings'
    CreditCard = 'CreditCard'
    Cash = 'Cash'
    LineOfCredit = 'LineOfCredit'
    Paypal = 'Paypal'
    MerchantAccount = 'MerchantAccount'
    InvestmentAccount = 'InvestmentAccount'
    Mortgage = 'Mortgage'
    OtherAsset = 'OtherAsset'
    OtherLiability = 'OtherLiability'


on_budget_dict = dict(
    undef=None,
    Checking=True,
    Savings=True,
    CreditCard=True,
    Cash=True,
    LineOfCredit=True,
    Paypal=True,
    MerchantAccount=True,
    InvestmentAccount=False,
    Mortgage=False,
    OtherAsset=False,
    OtherLiability=False,
)
on_budget_dict[None] = None


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Entity):
            pretreated = {}
            for namefield, field in obj.AllFields.items():
                value = getattr(obj, namefield)
                if value != undef:
                    pretreated[namefield] = field.pretreat(value)
            return pretreated
        elif isinstance(obj, ListofEntities):
            return list(obj)
        elif obj == undef:
            return
        else:
            return json.JSONEncoder.default(self, obj)


class UnknowEntityFieldValueError(Exception):
    pass


def obj_from_dict(obj_type, dictionary):
    treated = {}
    obt = obj_type()
    for key, value in dictionary.items():
        try:
            field = obt.AllFields[key]
        except KeyError:
            msg = 'Encountered field %s in a dictionary to create an entity of type %s, value %s ' % (key, obj_type,dictionary[key])
            logger.error(msg)
            raise UnknowEntityFieldValueError(msg)
        if isinstance(field, EntityField):
            treated[key] = field.posttreat(value)

    return obj_type(**treated)


ignored_fields_for_hash = ['id', 'credit_amount', 'cash_amount', 'feature_flags']


# adapted from http://stackoverflow.com/a/2954373/1685379
def addprop(inst, name, method, setter=None, cleaner=None):
    cls = type(inst)
    if not hasattr(cls, '__perinstance'):
        cls = type(cls.__name__, (cls,), {})
        cls.__perinstance = True
        inst.__class__ = cls
    p = property(method)
    setattr(cls, name, p)
    if setter is not None:
        setattr(cls, name, p.setter(setter))
    if cleaner is not None:
        setattr(cls, 'clean_' + name, cleaner)
    return p


class Entity(object):
    def __init__(self, *args, **kwargs):
        self.ListFields = {}
        self.Fields = {}
        self.AllFields = {}

        for namefield in dir(self):
            if namefield.startswith('__'): continue
            field = getattr(self, namefield)
            if isinstance(field, PropertyField):
                fieldc = copy.deepcopy(field)

                def getter(selfi):
                    if hasattr(selfi, '__prop_' + namefield):
                        return getattr(selfi, '__prop_' + namefield)
                    else:
                        return fieldc()(selfi)

                def setter(selfi, valuei):
                    setattr(selfi, '__prop_' + namefield, valuei)

                def cleaner(selfi):
                    delattr(selfi, '__prop_' + namefield)

                p = addprop(self, namefield, getter, setter, cleaner)
                if kwargs.get(namefield):
                    setattr(self.__class__, namefield, kwargs.get(namefield))
            elif isinstance(field, EntityField):
                self.Fields[namefield] = field
            elif isinstance(field, EntityListField):
                self.ListFields[namefield] = field
            else:
                continue
            self.AllFields[namefield] = field
            if isinstance(field, PropertyField): continue
            value = kwargs.get(namefield) if kwargs.get(namefield) is not None else field()
            setattr(self, namefield, value)
        super(Entity, self).__init__()

    id = EntityField(KeyGenerator.generateuuid)

    def __hash__(self):
        return self._hash()

    def _hash(self):
        t = tuple((k, v) for k, v in self.getdict().items() if k not in ignored_fields_for_hash)
        try:
            return hash(frozenset(t))
        except TypeError:
            pass

    def __str__(self):
        return self.getdict().__str__()

    def __repr__(self):
        return self.getdict().__repr__()

    def __unicode__(self):
        return self.__str__()

    def getdict(self):
        return {namefield: getattr(self, namefield) for namefield in self.AllFields}

    def get_changed_entities(self):
        firstrun = {namefield: getattr(self, namefield).get_changed_entities() for namefield in self.ListFields}
        return {namefield: value for namefield, value in firstrun.items() if value != []}

    def update_from_changed_entities(self, changed_entities):
        for namefield in self.ListFields:
            getattr(self, namefield).update_from_changed_entities(changed_entities.get(namefield))

    def update_from_dict(self, d):
        self.__dict__.update(d)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if isinstance(other, Entity):
            return self.getdict() == other.getdict()
        else:
            return False


class ListofEntities(list):
    def __init__(self, typ):
        super(ListofEntities, self).__init__()
        from collections import OrderedDict
        self._dict_entities = OrderedDict()
        self._dict_entities_hash = {}
        self.typeItems = typ
        self.type_instance = typ()
        self.changed = []

    def _update_hashes(self):
        self._dict_entities_hash = {hash(v): v for k, v in self._dict_entities.items()}

    def __hash__(self):
        return hash(frozenset(self._dict_entities))

    def __str__(self):
        return self._dict_entities.__str__()

    def __repr__(self):
        return self._dict_entities.__repr__()

    def __unicode__(self):
        return self.__str__()

    def __getitem__(self, item):
        return self._dict_entities.values().__getitem__(item)

    def update_from_changed_entities(self, changed_entities):
        if changed_entities is None:
            return
        for entity in changed_entities:
            if hasattr(entity, 'is_tombstone') and entity.is_tombstone:
                continue
            try:
                self._dict_entities[entity.id].update_from_changed_entities(entity)
            except KeyError:
                self._dict_entities[entity.id] = entity
        self._update_hashes()

    def get(self, entity_id):
        return self._dict_entities.get(entity_id)

    def extend(self, objects, track=True):
        if not all(isinstance(x, self.typeItems) for x in objects):
            raise ValueError('this ListofEntities can only contain %s' % self.typeItems.__name__)
        for o in objects:
            self._dict_entities[o.id] = o
            self._dict_entities_hash[hash(o)] = o
        if track:
            self.changed.extend(objects)

    def append(self, o, track=True):
        if not isinstance(o, self.typeItems):
            raise ValueError('this ListofEntities can only contain %s' % self.typeItems.__name__)
        self._dict_entities[o.id] = o
        self._dict_entities_hash[hash(o)] = o
        if track:
            self.changed.append(o)

    def delete(self, o, track=True):
        if not isinstance(o, self.typeItems):
            raise ValueError('this ListofEntities can only contain %s' % self.typeItems.__name__)
        if o.id in self._dict_entities:
            del self._dict_entities[o.id]
        if hash(o) in self._dict_entities_hash:
            del self._dict_entities_hash[hash(o)]
        if track:
            o.is_tombstone = True
            self.changed.append(o)

    def modify(self, o, track=True):
        if not isinstance(o, self.typeItems):
            raise ValueError('this ListofEntities can only contain %s' % self.typeItems.__name__)
        if o.id in self._dict_entities:
            h = hash(self._dict_entities[o.id])
            if h in self._dict_entities_hash:
                del self._dict_entities_hash[h]
            self._dict_entities[o.id] = o
            self._dict_entities_hash[hash(o)] = o
            if track:
                self.changed.append(o)

    def __iter__(self):
        return self._dict_entities.values().__iter__()

    def __len__(self):
        return len(self._dict_entities)

    def containsduplicate(self, item):
        if not isinstance(item, self.typeItems):
            return False
        else:
            return item._hash() in self._dict_entities_hash

    def __contains__(self, item):
        if not isinstance(item, self.typeItems):
            return False
        else:
            return item.id in self._dict_entities

    def get_changed_entities(self):
        return self.changed



from datetime import datetime




class EntityField(object):
    def pretreat(self, x):
        return x

    def posttreat(self, x):
        return x

    def __init__(self, default):
        self.default=default

    def __call__(self, *args, **kwargs):
        if callable(self.default):
            return self.default()
        else:
            return self.default


class EntityListField(object):
    def pretreat(self, x):
        return x

    def posttreat(self, x):
        return x

    def __init__(self, typearg):
        self.type = typearg

    def __call__(self, *args, **kwargs):
        from pynYNAB.Entity import ListofEntities
        return ListofEntities(self.type)


class AmountField(EntityField):
    def __init__(self):
        super(AmountField, self).__init__(0)

    def pretreat(self, x):
        return int(x * 1000) if x is not None else x

    def posttreat(self, x):
        return float(x) / 1000 if x is not None else x


class PropertyField(EntityField):
    def __init__(self, lambdafun):
        self.lambdafun = lambdafun

    def __call__(self, *args, **kwargs):
        return self.lambdafun

class DateField(EntityField):
    def pretreat(self, x):
        return x.strftime('%Y-%m-%d') if x is not None else x

    def posttreat(self, x):
        try:
            return datetime.strptime(x, '%Y-%m-%d').date() if x is not None else x
        except ValueError:
            pass


class DatesField(EntityField):
    d = DateField(None)

    def hash(self,value):
        return hash(frozenset(value))

    def pretreat(self, x):
        try:
            return [self.d.pretreat(i) for i in x]
        except AttributeError:
            from pynYNAB.Entity import AccountTypes
            if x in AccountTypes:
                return x

    def posttreat(self, x):
        try:
            return [self.d.posttreat(i) for i in x]
        except:
            # nYNAB servers can accept an account with account_type not set to a valid value apparently
            return None

class AccountTypeField(EntityField):
    def pretreat(self, x):
        try:
            return x.name
        except AttributeError:
            from pynYNAB.Entity import AccountTypes
            if x in AccountTypes:
                return x

    def posttreat(self, x):
        from pynYNAB.Entity import AccountTypes
        try:
            return getattr(AccountTypes, x)
        except:
            # nYNAB servers can accept an account with account_type not set to a valid value apparently
            return None



import time


def RateLimited(maxpersecond):
    minInterval = 1.0 / float(maxpersecond)
    def decorate(func):
        lastTimeCalled = [0.0]
        def rateLimitedFunction(*args,**kargs):
            elapsed = time.clock() - lastTimeCalled[0]
            leftToWait = minInterval - elapsed
            if leftToWait>0:
                time.sleep(leftToWait)
            ret = func(*args,**kargs)
            lastTimeCalled[0] = time.clock()
            return ret
        return rateLimitedFunction
    return decorate

def chunk(iterable, chunk_size):
    """Generate sequences of `chunk_size` elements from `iterable`."""
    iterable = iter(iterable)
    while True:
        chunk = []
        try:
            for _ in range(chunk_size):
                chunk.append(next(iterable))
            yield chunk
        except StopIteration:
            if chunk:
                yield chunk
            break

# http://stackoverflow.com/q/10480806/1685379
def equal_dicts(a, b, ignore_keys):
    ka = set(a).difference(ignore_keys)
    kb = set(b).difference(ignore_keys)
    return ka == kb and all(a[k] == b[k] for k in ka)


from pynYNAB.Entity import Entity, ListofEntities, obj_from_dict
from pynYNAB.schema.Fields import EntityListField
from pynYNAB.schema.budget import MasterCategory, Setting, MonthlyBudgetCalculation, AccountMapping, Subtransaction, \
    ScheduledSubtransaction, Subcategory, PayeeLocation, AccountCalculation, MonthlyAccountCalculation, \
    MonthlySubcategoryBudgetCalculation, ScheduledTransaction, Payee, MonthlySubcategoryBudget, PayeeRenameCondition, \
    Account, MonthlyBudget, TransactionGroup, DateField
from pynYNAB.schema.budget import Transaction
from pynYNAB.schema.catalog import UserBudget, UserSetting, BudgetVersion, User, CatalogBudget
from pynYNAB.scripts.config import get_logger


def knowledge_change(changed_entities):
    return sum(map(lambda v: len(v), [changed_entitity for dictkey, changed_entitity in changed_entities.items()]))


class Root(Entity):
    def __init__(self):

        self.logger = get_logger()
        self.current_device_knowledge = 0
        self.device_knowledge_of_server = 0
        super(Root, self).__init__()

    def sync(self, connection, opname):
        self.logger.debug('Starting sync')
        self.logger.debug('current_device_knowledge %s'%self.current_device_knowledge)
        self.logger.debug('device_knowledge_of_server %s' % self.device_knowledge_of_server)
        request_data = self.get_request_data()
        self.logger.debug('request_data starting_device_knowledge %s' % request_data['starting_device_knowledge'])
        self.logger.debug('request_data ending_device_knowledge %s' % request_data['ending_device_knowledge'])
        self.logger.debug('request_data device_knowledge_of_server %s' % request_data['device_knowledge_of_server'])
        self.logger.debug(request_data)
        syncData = connection.dorequest(request_data, opname)
        for namefield in self.ListFields:
            getattr(self, namefield).changed = []
        changed_entities = {}
        for name, value in syncData['changed_entities'].items():
            if isinstance(value, list):
                for entityDict in value:
                    obj = obj_from_dict(self.ListFields[name].type, entityDict)
                    try:
                        changed_entities[name].append(obj)
                    except KeyError:
                        changed_entities[name] = [obj]
            else:
                changed_entities[name] = self.AllFields[name].posttreat(value)


        server_knowledge_of_device=syncData['server_knowledge_of_device']
        self.logger.debug('from server server_knowledge_of_device %s'%server_knowledge_of_device)
        current_server_knowledge=syncData['current_server_knowledge']
        self.logger.debug('from server current_server_knowledge %s' % current_server_knowledge)

        change = current_server_knowledge - self.device_knowledge_of_server
        if change>0:
            self.logger.debug('Server knowledge has gone up by ' + str(change) + '. We should be getting back some entities from the server')
        if self.current_device_knowledge < server_knowledge_of_device:
            if self.current_device_knowledge!=0:
                self.logger.error('The server knows more about this device than we know about ourselves')
            self.current_device_knowledge = server_knowledge_of_device
        self.update_from_changed_entities(changed_entities)

        self.device_knowledge_of_server = current_server_knowledge

        self.logger.debug('Ending sync')
        self.logger.debug('current_device_knowledge %s' % self.current_device_knowledge)
        self.logger.debug('device_knowledge_of_server %s' % self.device_knowledge_of_server)
        pass

    def get_request_data(self):
        changed_entities = self.get_changed_entities()
        dictionary = {"starting_device_knowledge": self.current_device_knowledge,
                      "device_knowledge_of_server": self.device_knowledge_of_server,
                      "changed_entities": changed_entities}
        if any(changed_entities):
            dictionary['ending_device_knowledge'] = self.current_device_knowledge + 1
        else:
            dictionary['ending_device_knowledge'] = self.current_device_knowledge
        return dictionary


class Budget(Root):
    def __init__(self):
        super(Budget, self).__init__()
        self.budget_version_id = None

    be_transactions=EntityListField(Transaction)
    be_master_categories=EntityListField(MasterCategory)
    be_settings=EntityListField(Setting)
    be_monthly_budget_calculations=EntityListField(MonthlyBudgetCalculation)
    be_account_mappings=EntityListField(AccountMapping)
    be_subtransactions=EntityListField(Subtransaction)
    be_scheduled_subtransactions=EntityListField(ScheduledSubtransaction)
    be_monthly_budgets=EntityListField(MonthlyBudget)
    be_subcategories=EntityListField(Subcategory)
    be_payee_locations=EntityListField(PayeeLocation)
    be_account_calculations=EntityListField(AccountCalculation)
    be_monthly_account_calculations=EntityListField(MonthlyAccountCalculation)
    be_monthly_subcategory_budget_calculations=EntityListField(MonthlySubcategoryBudgetCalculation)
    be_scheduled_transactions=EntityListField(ScheduledTransaction)
    be_payees=EntityListField(Payee)
    be_monthly_subcategory_budgets=EntityListField(MonthlySubcategoryBudget)
    be_payee_rename_conditions=EntityListField(PayeeRenameCondition)
    be_accounts=EntityListField(Account)
    last_month=DateField(None)
    first_month=DateField(None)

    def get_request_data(self):
        request_data = super(Budget, self).get_request_data()
        request_data['budget_version_id'] = self.budget_version_id
        request_data['calculated_entities_included'] = False
        return request_data

    def get_changed_entities(self):
        changed_entities = super(Budget, self).get_changed_entities()
        if 'be_transactions' in changed_entities:
            changed_entities['be_transaction_groups'] = ListofEntities(TransactionGroup)
            for tr in changed_entities.pop('be_transactions'):
                subtransactions = ListofEntities(Subtransaction)
                if 'be_subtransactions' in changed_entities:
                    for subtr in [subtransaction for subtransaction in changed_entities.get('be_subtransactions') if
                                  subtransaction.entities_transaction_id == tr.id]:
                        changed_entities['be_subtransactions'].remove(subtr)
                        subtransactions.append(subtr)
                changed_entities['be_transaction_groups'].append(TransactionGroup(
                    id=tr.id,
                    be_transaction=tr,
                    be_subtransactions=subtransactions
                ))
        if changed_entities.get('be_subtransactions') is not None:
            del changed_entities['be_subtransactions']
        return changed_entities


class Catalog(Root):
    ce_user_budgets=EntityListField(UserBudget)
    ce_user_settings=EntityListField(UserSetting)
    ce_budget_versions=EntityListField(BudgetVersion)
    ce_users=EntityListField(User)
    ce_budgets=EntityListField(CatalogBudget)



import random
import unittest
from datetime import datetime, timedelta

from pynYNAB import KeyGenerator
from pynYNAB.Entity import AccountTypes
from pynYNAB.schema.budget import Transaction, Account, Subtransaction, Payee
from test_live.common_Live import commonLive
from test_live.common_Live import needs_account,needs_transaction


class liveTests(commonLive):
    def test_add_delete_budget(self):
        budget_name=KeyGenerator.generateuuid()
        self.client.create_budget(budget_name)
        self.reload()
        matches=[b for b in self.client.catalog.ce_budgets if b.budget_name==budget_name]
        self.assertTrue(len(matches) == 1)
        self.client.delete_budget(budget_name)
        self.reload()
        matches=[b for b in self.client.catalog.ce_budgets if b.budget_name==budget_name]
        self.assertTrue(len(matches) == 0)
        self.reload()

    def test_add_delete_account(self):
        account_type=AccountTypes.Checking
        account_name = KeyGenerator.generateuuid()
        budget = self.client.budget

        for account in budget.be_accounts:
            if account.account_name == account_name:
                return
        if len(budget.be_accounts) > 0:
            sortable_index = max(account.sortable_index for account in budget.be_accounts)
        else:
            sortable_index = 0

        account = Account(
            account_type=account_type,
            account_name=account_name,
            sortable_index=sortable_index,
        )

        self.client.add_account(account, balance=random.randint(-10, 10), balance_date=datetime.now())

        self.reload()

        self.assertIn(account, self.client.budget.be_accounts)

        self.client.delete_account(account)

        self.reload()

        result = self.client.budget.be_transactions.get(account.id)

        self.assertTrue(result is None)

    @needs_account()
    def test_add_deletetransaction(self):
        from datetime import datetime
        transaction = Transaction(
            amount=1,
            cleared='Uncleared',
            date=datetime.now(),
            entities_account_id=self.account.id,
        )
        self.client.add_transaction(transaction)

        self.reload()

        self.assertIn(transaction, self.client.budget.be_transactions)
        self.client.delete_transaction(transaction)
        self.reload()

        result = self.client.budget.be_transactions.get(transaction.id)
        self.assertTrue(result is None)

    @needs_account('account1')
    @needs_account('account2')
    def test_add_transfer(self):
        account1 = next(account for account in self.client.budget.be_accounts if account.account_name == 'account1')
        account2 = next(account for account in self.client.budget.be_accounts if account.account_name == 'account2')

        try:
            payee_2= next(payee for payee in self.client.budget.be_payees if payee.entities_account_id==account2.id)
        except StopIteration:
            payee_2=Payee(entities_account_id=account2.id)
            self.client.budget.be_payees.append(payee_2)
        try:
            payee_1= next(payee for payee in self.client.budget.be_payees if payee.entities_account_id==account1.id)
        except StopIteration:
            payee_1=Payee(entities_account_id=account1.id)
            self.client.budget.be_payees.append(payee_1)

        transaction1=Transaction(
            amount=random.randint(-10, 10),
            date=datetime.now(),
            entities_account_id=account1.id,
            entities_payee_id=payee_2.id
        )
        transaction2 = Transaction(
            amount=-transaction1.amount,
            date=datetime.now(),
            entities_account_id=account2.id,
            entities_payee_id=payee_1.id
        )
        self.client.budget.be_transactions.append(transaction1)
        self.client.budget.be_transactions.append(transaction2)
        self.client.sync()
        self.reload()
        self.assertIn(transaction1 , self.client.budget.be_transactions)
        self.assertIn(transaction2 , self.client.budget.be_transactions)

    @needs_account()
    def test_add_deletetransactions(self):
        from datetime import datetime

        transactions = [
            Transaction(
                amount=random.randint(-10, 10),
                cleared='Uncleared',
                date=datetime.now() - 8 * timedelta(days=365),
                entities_account_id=self.account.id,
            ), Transaction(
                amount=random.randint(-10, 10),
                cleared='Uncleared',
                date=datetime.now() + 8 * timedelta(days=365),
                entities_account_id=self.account.id,
            ), Transaction(
                amount=random.randint(-10, 10),
                cleared='Uncleared',
                date=datetime.now(),
                entities_account_id=self.account.id,
            )
        ]

        self.client.add_transactions(transactions)
        print('Time for request: ' + str(self.client.connection.lastrequest_elapsed.total_seconds()) + ' s')

        self.reload()
        for transaction in transactions:
            self.assertIn(transaction, self.client.budget.be_transactions)

        for transaction in transactions:
            self.client.delete_transaction(transaction)
        self.reload()

        for transaction in transactions:
            resulttransaction = self.client.budget.be_transactions.get(transaction.id)
            self.assertTrue(resulttransaction is None)

    @needs_account()
    def test_add_splittransactions(self):
        subcatsplit_id = next(subcategory.id for subcategory in self.client.budget.be_subcategories if
                              subcategory.internal_name == 'Category/__Split__')
        transaction = Transaction(
            amount=1,
            date=datetime.now(),
            entities_account_id=self.account.id,
            entities_subcategory_id=subcatsplit_id
        )
        sub1 = Subtransaction(
            amount=5000,
            entities_transaction_id=transaction.id
        )
        sub2 = Subtransaction(
            amount=5000,
            entities_transaction_id=transaction.id
        )
        self.client.budget.be_transactions.append(transaction)
        self.client.budget.be_subtransactions.append(sub1)
        self.client.budget.be_subtransactions.append(sub2)

        self.client.sync()

        self.reload()

        self.assertIn(transaction, self.client.budget.be_transactions)
        self.assertIn(sub1, self.client.budget.be_subtransactions)
        self.assertIn(sub2, self.client.budget.be_subtransactions)

    @needs_account()
    @needs_transaction
    def test_split(self):
        subcat1, subcat2 = tuple(random.sample(list(self.client.budget.be_subcategories), 2))
        subcatsplit_id = next(subcategory.id for subcategory in self.client.budget.be_subcategories if
                              subcategory.internal_name == 'Category/__Split__')
        self.transaction.entities_subcategory_id = subcatsplit_id

        sub1 = Subtransaction(
            amount=self.transaction.amount - 5000,
            entities_transaction_id=self.transaction.id,
            entities_subcategory_id=subcat1.id
        )
        sub2 = Subtransaction(
            amount=5000,
            entities_transaction_id=self.transaction.id,
            entities_subcategory_id=subcat2.id
        )

        self.client.budget.be_subtransactions.append(sub1)
        self.client.budget.be_subtransactions.append(sub2)
        self.client.budget.be_transactions.modify(self.transaction)

        self.client.sync()

        self.reload()

        self.assertIn(sub1, self.client.budget.be_subtransactions)
        self.assertIn(self.transaction, self.client.budget.be_transactions)
        self.assertIn(sub2, self.client.budget.be_subtransactions)



import logging
from functools import wraps

from pynYNAB.connection import NYnabConnectionError, nYnabConnection
from pynYNAB.roots import Budget, Catalog
from pynYNAB.schema.budget import Payee, Transaction
from pynYNAB.schema.catalog import BudgetVersion
from pynYNAB.scripts.config import get_logger
from pynYNAB.utils import chunk

logger = logging.getLogger('pynYNAB')


def clientfromargs(args, reset=False):
    connection = nYnabConnection(args.email, args.password)
    try:
        client = nYnabClient(connection, budget_name=args.budgetname)
        if reset:
            # deletes the budget
            client.delete_budget(args.budgetname)
            client.create_budget(args.budgetname)
            client.select_budget(args.budgetname)
        return client
    except BudgetNotFound:
        print('No budget by the name %s found in nYNAB'%args.budgetname)
        exit(-1)


class BudgetNotFound(Exception):
    pass


class nYnabClient(object):
    def __init__(self, nynabconnection, budget_name):
        self.logger = get_logger()
        if budget_name is None:
            logger.error('No budget name was provided')
            exit(-1)
        self.budget_name = budget_name
        self.connection = nynabconnection
        self.budget_name = budget_name
        self.catalog = Catalog()
        self.budget = Budget()
        self.budget_version = BudgetVersion()
        self.sync()

    def getinitialdata(self):
        try:
            getinitialdata = self.connection.dorequest({"device_info": {'id': self.connection.id}},
                                                       'getInitialUserData')
            self.budget.update_from_changed_entities(getinitialdata['budget'])
            self.budget_version.update_from_dict(getinitialdata['budget_version'])
            pass
        except NYnabConnectionError:
            pass

    def sync(self):
        # ending-starting represents the number of modifications that have been done to the data ?
        self.logger.debug('Catalog sync')
        self.catalog.sync(self.connection, 'syncCatalogData')
        if self.budget.budget_version_id is None:
            for catalogbudget in self.catalog.ce_budgets:
                if catalogbudget.budget_name == self.budget_name:
                    for budget_version in self.catalog.ce_budget_versions:
                        if budget_version.budget_id == catalogbudget.id:
                            self.budget.budget_version_id = budget_version.id
        if self.budget.budget_version_id is None and self.budget_name is not None:
            raise BudgetNotFound()
        else:
            self.logger.debug('Budget sync')
            self.budget.sync(self.connection, 'syncBudgetData')

    def operation(fn):
        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            fn(self, *args, **kwargs)
            self.sync()

        return wrapped

    @operation
    def add_account(self, account, balance, balance_date):
        payee = Payee(
            entities_account_id=account.id,
            enabled=True,
            auto_fill_subcategory_enabled=True,
            auto_fill_memo_enabled=False,
            auto_fill_amount_enabled=False,
            rename_on_import_enabled=False,
            name="Transfer : %s" % account.account_name
        )
        immediateincomeid = next(
            s.id for s in self.budget.be_subcategories if s.internal_name == 'Category/__ImmediateIncome__')
        startingbalanceid = next(p.id for p in self.budget.be_payees if p.internal_name == 'StartingBalancePayee')

        transaction = Transaction(
            accepted=True,
            amount=balance,
            entities_subcategory_id=immediateincomeid,
            cash_amount=0,
            cleared='Cleared',
            date=balance_date,
            entities_account_id=account.id,
            credit_amount=0,
            entities_payee_id=startingbalanceid,
            is_tombstone=False
        )

        self.budget.be_accounts.append(account)
        self.budget.be_payees.append(payee)
        self.budget.be_transactions.append(transaction)

    @operation
    def delete_account(self, account):
        self.budget.be_accounts.delete(account)

    @operation
    def add_transaction(self, transaction):
        self.budget.be_transactions.append(transaction)

    def add_transactions(self, transaction_list):
        for chunkelement in chunk(transaction_list, 50):
            self._add_transactions(chunkelement)

    @operation
    def _add_transactions(self, transaction_list):
        for transaction in transaction_list:
            self.budget.be_transactions.append(transaction)

    @operation
    def delete_transaction(self, transaction):
        self.budget.be_transactions.delete(transaction)

    def select_account_ui(self,create=False):
        accounts=list(self.budget.be_accounts)

        iaccount=0
        if create:
            print('#0 ###CREATE')
            iaccount=1

        for  account in accounts:
            print('#%d %s' % (iaccount, account.account_name))
            iaccount += 1
        if create:
            accounts=[None]+accounts


        while True:
            accountnumber = input('Which account? ')
            try:
                accountnumber = int(accountnumber)
                if 0 <= accountnumber <= len(accounts) - 1:
                    break
            except ValueError:
                pass
            print('Please enter a number between %d and %d' % (0, len(accounts) - 1))
            return accounts[accountnumber]

    @operation
    def delete_budget(self, budget_name):
        for budget in self.catalog.ce_budgets:
            if budget.budget_name == budget_name:
                budget.is_tombstone = True
                self.catalog.ce_budgets.modify(budget)

    def select_budget(self, budget_name):
        self.logger.debug('Catalog sync')
        self.catalog.sync(self.connection, 'syncCatalogData')
        for budget_version in self.catalog.ce_budget_versions:
            budget = self.catalog.ce_budgets.get(budget_version.budget_id)
            if budget.budget_name == budget_name:
                self.budget.budget_version_id = budget_version.id
                self.logger.debug('Budget sync')
                self.sync()
                break

    def create_budget(self, budget_name):
        import json
        currency_format = dict(
            iso_code='USD',
            example_format='123,456.78',
            decimal_digits=2,
            decimal_separator='.',
            symbol_first=True,
            group_separator=',',
            currency_symbol='$',
            display_symbol=True
        )
        date_format = dict(
            format='MM/DD/YYYY'
        )
        self.connection.dorequest(opname='CreateNewBudget',
                                  request_dic={
                                      "budget_name": budget_name,
                                      "currency_format": json.dumps(currency_format),
                                      "date_format": json.dumps(date_format)
                                  })

    @operation
    def clean_transactions(self):
        for transaction in self.budget.be_transactions:
            self.budget.be_transactions.delete(transaction)
        for subtransaction in self.budget.be_subtransactions:
            self.budget.be_subtransactions.delete(subtransaction)

    @operation
    def clean_budget(self):
        self.clean_transactions()
        for sub_category in [sub_category for sub_category in self.budget.be_subcategories if
                             sub_category.internal_name is None]:
            self.budget.be_subcategories.delete(sub_category)
        for mastercategory in [mastercategory for mastercategory in self.budget.be_master_categories if
                               mastercategory.deletable]:
            self.budget.be_master_categories.delete(mastercategory)
        self.clean_transactions()
        for payee in [payee for payee in self.budget.be_payees if payee.internal_name is None]:
            self.budget.be_payees.delete(payee)
        for account in self.budget.be_accounts:
            self.budget.be_accounts.delete(account)



import json
import unittest

from pynYNAB.Entity import Entity, ComplexEncoder, ListofEntities, undef, AccountTypes, addprop
from pynYNAB.roots import Budget, Catalog
from pynYNAB.schema.Fields import EntityField, EntityListField, DateField, PropertyField
from pynYNAB.schema.budget import Account, AccountCalculation, AccountMapping, MasterCategory, Transaction, Subcategory, \
    MonthlyAccountCalculation, MonthlyBudget, MonthlySubcategoryBudget, MonthlyBudgetCalculation, \
    MonthlySubcategoryBudgetCalculation, PayeeLocation, Payee, PayeeRenameCondition, ScheduledSubtransaction, \
    ScheduledTransaction, Setting, Subtransaction, TransactionGroup
from pynYNAB.schema.catalog import BudgetVersion, CatalogBudget, User, UserBudget, UserSetting


class Test1(unittest.TestCase):
    maxDiff = None

    def testEntityjson(self):
        class MyEntity(Entity):
            greatfield = EntityField(2)

        obj = MyEntity()
        jsonroundtrip = json.loads(json.dumps(obj, cls=ComplexEncoder))
        assert (jsonroundtrip == {'id': obj.id, 'greatfield': 2})

    def testequality(self):
        tr1 = Transaction()
        tr2 = Transaction()
        self.assertNotEqual(tr1, tr2)

        tr1 = Transaction(entities_account_id=1)
        tr2 = Transaction(entities_account_id=2)
        self.assertNotEqual(tr1, tr2)

    def testentityIn(self):
        tr1 = Transaction()
        transactions = ListofEntities(Transaction)
        transactions.append(tr1)
        self.assertIn(tr1, transactions)

    def testentityIn2(self):
        tr1 = Transaction()
        tr2 = Transaction()
        transactions = ListofEntities(Transaction)
        transactions.append(tr1)
        self.assertNotIn(tr2, transactions)

    def test_hash(self):
        tr1 = Transaction()
        result = tr1._hash()
        self.assertIsInstance(result, int)


    def testprop(self):

        namefield='p'
        default= lambda self:self.y
        def pgetter(self):
            if hasattr(self,'__prop_'+namefield):
                return getattr(self,'__prop_'+namefield)
            else:
                return default(self)

        def psetter(self,value):
            setattr(self,'__prop_'+namefield,value)

        class myClass(object):
            y = 1

        obj1=myClass()
        addprop(obj1,namefield,pgetter,psetter)

        self.assertEqual(obj1.p,1)
        obj1.y=3
        self.assertEqual(obj1.p,3)
        obj1.p=5
        self.assertEqual(obj1.p,5)
        obj1.y=7
        self.assertNotEqual(obj1.p,7)

    def test_lambdaprop(self):
        class MockEntity(Entity):
            input = EntityField(True),
            override = PropertyField(lambda x: x.input)
        # override behaves like a @property attribute:

        entity1=MockEntity()
        self.assertEqual(entity1.override,entity1.input)
        entity1.input=False
        self.assertEqual(entity1.override,entity1.input)
        entity1.input=12
        self.assertEqual(entity1.override,entity1.input)

        # unless set directly:

        entity1.override=42
        self.assertEqual(entity1.override,42)
        # then it behaves like a normal attribute:
        entity1.input=False
        self.assertEqual(entity1.override,42)

        # we can clear and come back to @property behavior
        entity1.clean_override()

        self.assertEqual(entity1.override,entity1.input)
        entity1.input=False
        self.assertEqual(entity1.override,entity1.input)
        entity1.input=12
        self.assertEqual(entity1.override,entity1.input)

    def testimports(self):
        types = [
            Account,
            AccountCalculation,
            AccountMapping,
            Budget,
            MasterCategory,
            MonthlyAccountCalculation,
            MonthlyBudget,
            MonthlyBudgetCalculation,
            MonthlySubcategoryBudget,
            MonthlySubcategoryBudgetCalculation,
            Payee,
            PayeeLocation,
            PayeeRenameCondition,
            ScheduledSubtransaction,
            ScheduledTransaction,
            Setting,
            Subcategory,
            Subtransaction,
            Transaction,
            TransactionGroup,
            BudgetVersion,
            Catalog,
            CatalogBudget,
            User,
            UserBudget,
            UserSetting
        ]

        def checkequal(l1, l2):
            return len(l1) == len(l2) and sorted(l1) == sorted(l2)

        for typ in types:
            obj = typ()
            self.assertIsInstance(obj.AllFields, dict)
            for f in obj.AllFields:
                self.assertTrue(isinstance(obj.AllFields[f], EntityField) or isinstance(obj.AllFields[f], EntityListField))
            self.assertTrue(checkequal(obj.getdict().keys(), obj.AllFields.keys()))

            valuesleft=list(obj.getdict().values())
            valuesright=[getattr(obj, f) for f in obj.AllFields.keys()]

            unhashableleft = [v for v in valuesleft if v.__hash__ is None]
            hashableleft = [v for v in valuesleft if v.__hash__ is not None]

            unhashableright = [v for v in valuesright if v.__hash__ is None]
            hashableright = [v for v in valuesright if v.__hash__ is not None]
            self.assertEqual(set(hashableleft),set(hashableright))
            self.assertEqual(unhashableleft,unhashableright)

    def testupdatechangedentities(self):
        obj = Budget()
        assert (obj.be_accounts.__class__ == ListofEntities)
        assert (obj.be_accounts.typeItems == Account)
        assert (len(obj.be_accounts) == 0)
        account = Account()
        changed_entities = dict(
            be_accounts=[account]
        )
        obj.update_from_changed_entities(changed_entities)
        assert (len(obj.be_accounts) == 1)
        assert (next(obj.be_accounts.__iter__()) == account)

    def testappend(self):

        obj = Budget()
        account = Account(None)
        obj.be_accounts.append(account)
        assert (len(obj.be_accounts) == 1)
        assert (list(obj.be_accounts)[-1] == account)

    def testappendBad(self):
        obj = Budget()
        transaction = Transaction()
        self.assertRaises(ValueError, lambda: obj.be_accounts.append(transaction))

    def testCE_nochange(self):
        obj = Transaction(None)
        self.assertEqual(obj.get_changed_entities(), {})

    def testCE_simpleadd(self):
        obj = Budget()
        account = Account()
        obj.be_accounts.append(account)
        self.assertEqual(obj.get_changed_entities(), {'be_accounts': [account]})

    def testCE_simpledelete(self):
        obj = Budget()
        account = Account()
        obj.be_accounts.delete(account)
        self.assertEqual(obj.get_changed_entities(), {'be_accounts': [account]})

    def testCE_simplechange(self):
        obj = Budget()
        account1 = Account()
        obj.be_accounts.append(account1)
        account2 = Account(id=account1.id, account_name='BLA')
        obj.be_accounts.changed = []
        obj.be_accounts.modify(account2)
        self.assertEqual(obj.get_changed_entities(), {'be_accounts': [account2]})

    def test_str(self):
        # tests no exceptions when getting the string representation of some entities
        obj = Transaction()
        obj.__str__()
        obj.__unicode__()

        obj2 = Budget()
        obj2.be_accounts.__str__()
        obj2.be_accounts.__unicode__()

    def jsondefault(self):
        encoded = json.dumps('test', cls=ComplexEncoder)
        self.assertEqual(encoded, 'test')



import configargparse
import random
import unittest
from datetime import datetime

from pynYNAB import KeyGenerator
from pynYNAB.Client import clientfromargs
from pynYNAB.Entity import AccountTypes
from pynYNAB.schema.budget import Account, Payee
from pynYNAB.schema.budget import Transaction

from pynYNAB.scripts.config import parser
from functools import wraps


def needs_account(account_name=None):
    def decorator(fn):
        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            for account in self.client.budget.be_accounts:
                if account_name is None or account.account_name == account_name:
                    self.account = account
                    fn(self, *args, **kwargs)
                    return
            self.account = self.util_add_account(account_name)
            fn(self, *args, **kwargs)
            return
        return wrapped
    return decorator


def needs_transaction(fn):
    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        for transaction in self.client.budget.be_transactions:
            if transaction.entities_account_id == self.account.id:
                self.transaction = transaction
                fn(self, *args, **kwargs)
                return
        self.util_add_transaction()
        raise ValueError('No available transaction !')

    return wrapped

class commonLive(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(commonLive, self).__init__(*args, **kwargs)
        self.account = None
        self.budget = None
        self.transaction = None
        self.client = None

    def reload(self):
        # parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        self.client = clientfromargs(args)

    def setUp(self):
        self.reload()

    def util_add_account(self,account_name=None):
        if account_name is None:
            account_name = KeyGenerator.generateuuid()
        account = Account(
            account_type=random.choice(list(AccountTypes)),
            account_name= account_name
        )

        self.client.add_account(account, balance=random.randint(-10, 10), balance_date=datetime.now())
        return account



    def util_add_transaction(self):
        transaction = Transaction(
            amount=1,
            cleared='Uncleared',
            date=datetime.now(),
            entities_account_id=self.account.id,
        )
        self.client.add_transaction(transaction)

    def util_get_empty_account_by_name_if_doesnt_exist(self, name):
        accounts = {a.account_name: a for a in self.client.budget.be_accounts if
                    a.account_name == name}
        if name in accounts:
            account = accounts[name]
            self.client.delete_account(account)

        account = Account(
            account_type=AccountTypes.Checking,
            account_name=name
        )

        self.client.add_account(account, balance=random.randint(-10, 10), balance_date=datetime.now())
        self.reload()
        self.assertIn(account, self.client.budget.be_accounts)
        return account

    def util_add_payee_by_name_if_doesnt_exist(self, name):
        payees = {p.name: p for p in self.client.budget.be_payees if
                  p.name == name}
        if name in payees:
            return payees[name]
        payee = Payee(
            name=name
        )

        self.client.budget.be_payees.append(payee)
        self.client.sync()
        return payee




from pynYNAB.Entity import Entity, undef
from pynYNAB.schema.Fields import EntityField


class CatalogBudget(Entity):
    budget_name=EntityField(None)
    is_tombstone=EntityField(None)


class UserBudget(Entity):
    budget_id=EntityField(undef)
    user_id=EntityField(None)
    is_tombstone=EntityField(False)
    permissions=EntityField(None)


class UserSetting(Entity):
    setting_name=EntityField(None)
    user_id=EntityField(None)
    setting_value=EntityField(None)


class User(Entity):
    username=EntityField(None)
    trial_expires_on=EntityField(None)
    is_tombstone=EntityField(False)
    email=EntityField(None)
    feature_flags=EntityField(None)
    is_subscribed=EntityField(None)


class BudgetVersion(Entity):
    date_format=EntityField(None)
    last_accessed_on=EntityField(None)
    currency_format=EntityField(None)
    budget_id=EntityField(None)
    is_tombstone=EntityField(False)
    version_name=EntityField(None)
    source=EntityField(None)






import os
import subprocess

with open('README.rst', 'w') as readme:
    for module in os.listdir(os.path.dirname(__file__)):
        if module == os.path.basename(__file__) or module == '__init__.py' or module[-3:] != '.py':
            continue
        readme.write('\n')
        readme.writelines('\n'.join([module,'-'*len(module),'.. code-block:: ','']))
        process=subprocess.Popen([module,'-h'],shell=True,stdout=subprocess.PIPE)
        readme.write('\n')
        for line in process.stdout:
            readme.write('    '+line)

    readme.writelines("""
Command Line / Config File Arguments
====================================
Args that start with '--' (eg. --email) can also be set in the config file
(ynab.conf). The recognized syntax for setting (key, value) pairs is based
on the INI and YAML formats (e.g. key=value or foo=TRUE). For full
documentation of the differences from the standards please refer to the
ConfigArgParse documentation. If an arg is specified in more than one
place, then commandline values override config file values.
""")




# coding=utf-8
import json
import logging
import uuid
from time import sleep

import requests
from requests.cookies import RequestsCookieJar

from pynYNAB.Entity import ComplexEncoder
from pynYNAB.utils import RateLimited


class NYnabConnectionError(Exception):
    pass

requests.packages.urllib3.disable_warnings()

class nYnabConnection(object):
    url = 'https://app.youneedabudget.com/users/login'
    urlCatalog = 'https://app.youneedabudget.com/api/v1/catalog'

    def _init_session(self):
        self.session.cookies = RequestsCookieJar()

        self.session.headers['X-YNAB-Device-Id'] = self.id
        self.session.headers['User-Agent'] = 'python nYNAB API bot - rienafairefr rienafairefr@gmail.com'

        firstlogin = self.dorequest({"email": self.email, "password": self.password, "remember_me": True,
                                     "device_info": {"id": self.id}}, 'loginUser')
        if firstlogin is None:
            raise NYnabConnectionError('Couldnt connect with the provided email and password')
        self.sessionToken = firstlogin["session_token"]
        self.session.headers['X-Session-Token'] = self.sessionToken

    def __init__(self, email, password):
        self.email = email
        self.password = password
        self.session = requests.Session()
        self.sessionToken = None
        self.id = str(uuid.uuid3(uuid.NAMESPACE_DNS, 'rienafairefr.pynYNAB'))
        self.lastrequest_elapsed=None
        self.logger = logging.getLogger('pynYnab')
        self._init_session()

    @RateLimited(maxpersecond=5)
    def dorequest(self, request_dic, opname):
        """
        :param request_dic: a dictionary containing parameters for the request
        :param opname: An API operation name, available are: loginUser,getInitialUserData,logoutUser,createNewBudget,
        freshStartABudget,cloneBudget,deleteTombstonedBudgets,syncCatalogData,syncBudgetData,getInstitutionList,
        getInstitutionLoginFields,getInstitutionAccountList,registerAccountForDirectConnect,
        updateDirectConnectCredentials,poll,createFeedback,runSqlStatement
        :return: the dictionary of the result of the request
        """
        # Available operations :

        params = { u'operation_name': opname,'request_data': json.dumps(request_dic, cls=ComplexEncoder),}
        self.logger.debug('%s  ... %s ' % (opname,params))
        r = self.session.post(self.urlCatalog, params, verify=False)
        self.lastrequest_elapsed=r.elapsed
        js = r.json()
        if r.status_code != 200:
             self.logger.debug('non-200 HTTP code: %s ' % r.text)
        if js['error'] is None:
            return js
        else:
            error=js['error']
            if r.status_code == 500:
                raise NYnabConnectionError('Uunrecoverable server error, sorry YNAB')
            if error['id'] == 'user_not_found':
                 self.logger.error('API error, User Not Found')
            elif error['id'] == 'id=user_password_invalid':
                 self.logger.error('API error, User-Password combination invalid')
            elif error['id'] == 'request_throttled':
                 self.logger.debug('API Rrequest throttled')
                 retyrafter=r.headers['Retry-After']
                 self.logger.debug('Waiting for %s s' % retyrafter)
                 sleep(float(retyrafter))
                 return self.dorequest(request_dic,opname)
            else:
                raise NYnabConnectionError('Unknown Error \"%s\" was returned from the API'%error['id'])
















import os.path

from setuptools import setup, find_packages

# Get the long description from the relevant file
__here__ = os.path.dirname(os.path.realpath(__file__))
readme_filename = os.path.join(os.path.dirname(__file__), 'README.rst')
with open(readme_filename, 'r') as f:
  readme = f.read()
try:
    import pypandoc
    readme = pypandoc.convert(readme, to='rst', format='md')
except:
    print("Conversion of long_description from markdown to reStructuredText failed, skipping...")
    pass

setup(
    name='pynYNAB',
    version='0.2',
    # Note: change 'master' to the tag name when release a new verion
    download_url='https://github.com/rienafairefr/nYNABapi/tarball/master',

    description=('Library for working with the nYNAB private API'
                 'Budget/Transactions Read/Write/Migration tools'),

    long_description=readme,

    url='https://github.com/rienafairefr/nYNABapi',

    author='rienafairefr',
    author_email='rienafairefr@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Topic :: Office/Business',
        'Topic :: Office/Business :: Financial',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],

    keywords=['nynab', 'ynab'],

    packages=find_packages(),

    install_requires=[
        'configargparse',
        'ofxtools',
        'pynab',
        'requests',
        'jsontableschema',
        'unicodecsv',
        'appdirs'
    ],

    package_data={
        'pynYNAB': ['tests/*'],
    },
)



import json
import os
from datetime import datetime
from tempfile import gettempdir

import configargparse

from pynYNAB.Entity import ComplexEncoder
from pynYNAB.schema.budget import Transaction
from pynYNAB.scripts.csvimport import do_csvimport
from test_live.common_Live import commonLive
from test_live.test_live import needs_account


class TestCsv(commonLive):
    def getTr(self, date, payee, amount, memo, account):
        imported_date = datetime.now().date()
        return Transaction(
            entities_account_id=self.util_get_empty_account_by_name_if_doesnt_exist(account).id,
            date=date,
            entities_payee_id=self.util_add_payee_by_name_if_doesnt_exist(payee).id,
            imported_payee=payee,
            source='Imported',
            memo=memo,
            amount=amount,
            cash_amount=amount,
            imported_date=imported_date
        )

    @needs_account('Credit')
    def test_duplicate(self):
        parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        args.schema = 'example'
        args.csvfile = os.path.join(gettempdir(),'data.csv')
        args.accountname = None
        args.import_duplicates = False
        args.logginglevel = 'debug'

        content = """Date,Payee,Amount,Memo,Account
2016-02-01,Super Pants Inc.,-20,Buying pants,Credit
"""
        with open(args.csvfile, mode='w') as f:
            f.writelines(content)

        transaction=self.getTr(datetime(year=2016, month=2, day=1).date(), 'Super Pants Inc.', -20, 'Buying pants', 'Credit')

        for i in range(2):
            do_csvimport(args)
            self.reload()
            identical=[tr2 for tr2 in self.client.budget.be_transactions if transaction._hash() == tr2._hash()]
            print('Transactions with same hash: %s'%len(identical))
            self.assertTrue(len(identical) == 1)

    @needs_account('Cash')
    def test_duplicateForced(self):
        parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        args.schema = 'example'
        args.csvfile = os.path.join(gettempdir(),'data.csv')
        args.accountname = None
        args.import_duplicates = True
        args.logginglevel = 'debug'

        content = """Date,Payee,Amount,Memo,Account
2016-02-01,Super Pants Inc.,-20,Buying pants,Cash
"""
        with open(args.csvfile, mode='w') as f:
            f.writelines(content)

        transaction=self.getTr(datetime(year=2016, month=2, day=1).date(), 'Super Pants Inc.', -20, 'Buying pants', 'Cash')

        do_csvimport(args)
        self.reload()
        do_csvimport(args)
        self.reload()
        self.assertTrue(len([tr2 for tr2 in self.client.budget.be_transactions if transaction._hash() == tr2._hash()]) == 2)

    @needs_account('Cash')
    @needs_account('Checking Account')
    @needs_account('Savings')
    def test_import(self):
        parser = configargparse.getArgumentParser('pynYNAB')
        args = parser.parse_known_args()[0]
        args.schema = 'example'
        args.csvfile = os.path.join(gettempdir(),'data.csv')
        args.accountname = None
        args.import_duplicates=False
        args.logginglevel = 'debug'

        content = """Date,Payee,Amount,Memo,Account
2016-02-01,Super Pants Inc.,-20,Buying pants,Cash
2016-02-02,Thai Restaurant,-10,Food,Checking Account
2016-02-03,,10,Saving!,Savings
        """
        with open(args.csvfile, mode='w') as f:
            f.writelines(content)

        Transactions = [
            self.getTr(datetime(year=2016, month=2, day=1).date(), 'Super Pants Inc.', -20, 'Buying pants', 'Cash'),
            self.getTr(datetime(year=2016, month=2, day=2).date(), 'Thai Restaurant', -10, 'Food', 'Checking Account'),
            self.getTr(datetime(year=2016, month=2, day=3).date(), '', 10, 'Saving!', 'Savings'),
        ]

        do_csvimport(args)
        self.reload()
        for tr in Transactions:
            print(json.dumps(tr, cls=ComplexEncoder))
            print(json.dumps(
                [tr2 for tr2 in self.client.budget.be_transactions if tr2.amount == tr.amount],
                cls=ComplexEncoder))
            self.assertTrue(self.client.budget.be_transactions.containsduplicate(tr))



import argparse
import logging
import os
from appdirs import AppDirs
import configargparse

configfile = 'ynab.conf'
if not os.path.exists(configfile):
    myAppdir = AppDirs('pynYNAB').user_config_dir
    configfile = os.path.join(myAppdir, configfile)



parser = configargparse.getArgumentParser('pynYNAB',default_config_files=[configfile],
                                          add_env_var_help=True,
                                          add_config_file_help=True,
                                          auto_env_var_prefix='NYNAB_')
parser.add_argument('--email', metavar='Email', type=str, required=False,
                    help='The Email User ID for nYNAB')
parser.add_argument('--password', metavar='Password', type=str, required=False,
                    help='The Password for nYNAB')
parser.add_argument('--logginglevel', metavar='LoggingLevel', type=str.lower, required=False,default='error',
                    choices=['critical','error','warn','warning','info','debug'],
                    help='Logging Level')

parser.add_argument('--budgetname', metavar='BudgetName', type=str, required=False,
                        help='The nYNAB budget to use')


def get_logger(args=None):
    args = parser.parse_known_args()[0] if args is None else args
    logginglevel = args.logginglevel.upper()
    logger = logging.getLogger('pynYNAB')
    logging.basicConfig()
    logger.setLevel(logginglevel)
    return logger


def test_common_args(args):
    if args.email is None:
        get_logger().error('No email user ID provided, please specify it at the command line or in %s'%(configfile,))
        exit(-1)
    if args.password is None:
        get_logger().error('No password provided, please specify it at the command line or in %s'%(configfile,))
        exit(-1)
    if args.budgetname is None:
        get_logger().error('No budget name provided, please specify it at the command line or in %s'%(configfile,))
        exit(-1)

def alert_quit_badconfig():
    print('Please modify ynab.conf situated in %s' % myAppdir)
    import shutil

    shutil.copy('ynab.conf.format', os.path.join(myAppdir, 'ynab.conf'))
    exit(-1)



from pynYNAB.Entity import Entity, undef, on_budget_dict, AccountTypes
from pynYNAB.schema.Fields import EntityField, AmountField, DateField, AccountTypeField, EntityListField, DatesField, \
    PropertyField


class Transaction(Entity):
    accepted=EntityField(True)
    amount=AmountField()
    cash_amount=AmountField()
    check_number=EntityField(None)
    cleared=EntityField('Uncleared')
    credit_amount=AmountField()
    date=DateField(None)
    date_entered_from_schedule=DateField(None)
    entities_account_id=EntityField(None)
    entities_payee_id=EntityField(None)
    entities_scheduled_transaction_id=EntityField(None)
    entities_subcategory_id=EntityField(None)
    flag=EntityField("")
    imported_date=DateField(None)
    imported_payee=EntityField(None)
    is_tombstone=EntityField(False)
    matched_transaction_id=EntityField(None)
    memo=EntityField(None)
    source=EntityField(None)
    subcategory_credit_amount_preceding=EntityField(0)
    transfer_account_id=EntityField(None)
    transfer_subtransaction_id=EntityField(None)
    transfer_transaction_id=EntityField(None)
    ynab_id=EntityField(None)


class MasterCategory(Entity):
    deletable=EntityField(True)
    internal_name=EntityField(None)
    is_hidden=EntityField(False)
    is_tombstone=EntityField(False)
    name=EntityField(None)
    note=EntityField(None)
    sortable_index=EntityField(None)


class Setting(Entity):
    setting_name=EntityField(None)
    setting_value=EntityField(None)


class MonthlyBudgetCalculation(Entity):
    additional_to_be_budgeted=AmountField()
    age_of_money=EntityField(None)
    available_to_budget=EntityField(None)
    balance=EntityField(None)
    budgeted=EntityField(None)
    calculation_notes=EntityField(None)
    cash_outflows=AmountField()
    credit_outflows=AmountField()
    deferred_income=AmountField()
    entities_monthly_budget_id=EntityField(None)
    hidden_balance=AmountField()
    hidden_budgeted=AmountField()
    hidden_cash_outflows=AmountField()
    hidden_credit_outflows=AmountField()
    immediate_income=AmountField()
    is_tombstone=EntityField(False)
    over_spent=AmountField()
    previous_income=AmountField()
    uncategorized_balance=AmountField()
    uncategorized_cash_outflows=AmountField()
    uncategorized_credit_outflows=AmountField()


class AccountMapping(Entity):
    date_sequence=DateField(None)
    entities_account_id=EntityField(None)
    hash=EntityField(None)
    fid=EntityField(None)
    is_tombstone=EntityField(False)
    salt=EntityField(None)
    shortened_account_id=EntityField(None)
    should_flip_payees_memos=EntityField(None)
    should_import_memos=EntityField(None)
    skip_import=EntityField(None)


class Subtransaction(Entity):
    amount=AmountField()
    cash_amount=AmountField()
    check_number=EntityField(None)
    credit_amount=AmountField()
    entities_payee_id=EntityField(None)
    entities_subcategory_id=EntityField(None)
    entities_transaction_id=EntityField(None)
    is_tombstone=EntityField(False)
    memo=EntityField(None)
    sortable_index=EntityField(0)
    subcategory_credit_amount_preceding=EntityField(0)
    transfer_account_id=EntityField(None)
    transfer_transaction_id=EntityField(None)


class ScheduledSubtransaction(Entity):
    amount=AmountField()
    entities_payee_id=EntityField(None)
    entities_scheduled_transaction_id=EntityField(None)
    entities_subcategory_id=EntityField(None)
    is_tombstone=EntityField(False)
    memo=EntityField(None)
    sortable_index=EntityField(0)
    transfer_account_id=EntityField(None)


class MonthlyBudget(Entity):
    is_tombstone=EntityField(False)
    month=EntityField(None)
    note=EntityField(None)


class Subcategory(Entity):
    entities_account_id=EntityField(None)
    entities_master_category_id=EntityField(None)
    goal_creation_month=EntityField(None)
    goal_type=EntityField(None)
    internal_name=EntityField(None)
    is_hidden=EntityField(False)
    is_tombstone=EntityField(False)
    monthly_funding=EntityField(None)
    name=EntityField(None)
    note=EntityField(None)
    sortable_index=EntityField(0)
    target_balance=AmountField()
    target_balance_month=EntityField(None)
    type=EntityField(None)


class PayeeLocation(Entity):
    entities_payee_id=EntityField(None)
    is_tombstone=EntityField(False)
    latitude=EntityField(None)
    longitude=EntityField(None)


class AccountCalculation(Entity):
    cleared_balance=AmountField()
    entities_account_id=EntityField(None)
    error_count=EntityField(None)
    info_count=EntityField(None)
    is_tombstone=EntityField(False)
    transaction_count=EntityField(None)
    uncleared_balance=AmountField()
    warning_count=EntityField(None)


class MonthlyAccountCalculation(Entity):
    cleared_balance=AmountField()
    entities_account_id=EntityField(None)
    error_count=EntityField(None)
    info_count=EntityField(None)
    is_tombstone=EntityField(False)
    month=EntityField(None)
    transaction_count=EntityField(None)
    uncleared_balance=AmountField()
    warning_count=EntityField(None)
    rolling_balance=AmountField()


class MonthlySubcategoryBudgetCalculation(Entity):
    additional_to_be_budgeted=EntityField(None)
    all_spending=AmountField()
    all_spending_since_last_payment=AmountField()
    balance=AmountField()
    balance_previous_month=AmountField()
    budgeted_average=AmountField()
    budgeted_cash_outflows=AmountField()
    budgeted_credit_outflows=AmountField()
    budgeted_previous_month=AmountField()
    budgeted_spending=AmountField()
    cash_outflows=AmountField()
    credit_outflows=AmountField()
    entities_monthly_subcategory_budget_id=EntityField(None)
    goal_expected_completion=EntityField(None)
    goal_overall_funded=AmountField()
    goal_overall_left=AmountField()
    goal_percentage_complete=EntityField(None)
    goal_target=EntityField(None)
    goal_under_funded=EntityField(None)
    is_tombstone=EntityField(False)
    overspending_affects_buffer=EntityField(None)
    payment_average=AmountField()
    payment_previous_month=AmountField()
    spent_average=AmountField()
    spent_previous_month=AmountField()
    unbudgeted_cash_outflows=AmountField()
    unbudgeted_credit_outflows=AmountField()
    upcoming_transactions=AmountField()
    upcoming_transactions_count=EntityField(None)
    positive_cash_outflows=AmountField()


class ScheduledTransaction(Entity):
    amount=AmountField()
    date=DateField(None)
    entities_account_id=EntityField(None)
    entities_payee_id=EntityField(None)
    entities_subcategory_id=EntityField(None)
    flag=EntityField(None)
    frequency=EntityField(None)
    is_tombstone=EntityField(False)
    memo=EntityField(None)
    transfer_account_id=EntityField(None)
    upcoming_instances=DatesField([])


class Payee(Entity):
    auto_fill_amount=AmountField()
    auto_fill_amount_enabled=EntityField(None)
    auto_fill_memo=EntityField(None)
    auto_fill_memo_enabled=EntityField(None)
    auto_fill_subcategory_enabled=EntityField(None)
    auto_fill_subcategory_id=EntityField(None)
    enabled=EntityField(True)
    entities_account_id=EntityField(None)
    internal_name=EntityField(None)
    is_tombstone=EntityField(False)
    name=EntityField(None)
    rename_on_import_enabled=EntityField(None)


class MonthlySubcategoryBudget(Entity):
    budgeted=AmountField()
    entities_monthly_budget_id=EntityField(None)
    entities_subcategory_id=EntityField(None)
    is_tombstone=EntityField(False)
    note=EntityField(None)
    overspending_handling=EntityField(None)


class TransactionGroup(Entity):
    be_transaction=EntityField(None)
    be_subtransactions=EntityField(None)
    be_matched_transaction=EntityField(None)


class PayeeRenameCondition(Entity):
    entities_payee_id=EntityField(None)
    is_tombstone=EntityField(False)
    operand=EntityField(None)
    operator=EntityField(None)


def on_budget_default(self):
    return on_budget_dict[self.account_type.value]


class Account(Entity):
    account_name=EntityField(None)
    account_type=AccountTypeField(AccountTypes.undef)
    direct_connect_account_id=EntityField(undef)
    direct_connect_enabled=EntityField(False)
    direct_connect_institution_id=EntityField(undef)
    hidden=EntityField(False)
    is_tombstone=EntityField(False)
    last_entered_check_number=EntityField(None)
    last_imported_at=EntityField(undef)
    last_imported_error_code=EntityField(undef)
    last_reconciled_balance=EntityField(None)
    last_reconciled_date=DateField(None)
    direct_connect_last_error_code=EntityField(None)
    direct_connect_last_imported_at=DateField(None)
    note=EntityField(None)
    sortable_index=EntityField(0)
    on_budget=PropertyField(on_budget_default)



from pynYNAB.Client import clientfromargs
from pynYNAB.scripts.config import parser

# used to ping the nYNAB API to check that the sync works

args = parser.parse_known_args()[0]
client = clientfromargs(args)
client.sync()


import configargparse
import inspect
from datetime import datetime
from ofxtools import OFXTree

from pynYNAB.Client import clientfromargs
from pynYNAB.schema.budget import Transaction, Payee
from pynYNAB.scripts.config import test_common_args


def ofximport_main():
    print('pynYNAB OFX import')
    """Manually import an OFX into a nYNAB budget"""

    parser = configargparse.getArgumentParser('pynYNAB')
    parser.description = inspect.getdoc(ofximport_main)
    parser.add_argument('ofxfile', metavar='OFXPath', type=str,
                        help='The OFX file to import')

    args = parser.parse_args()
    test_common_args(args)
    do_ofximport(args)

def do_ofximport(args, client = None):
    if client is None:
        client = clientfromargs(args)

    tree = OFXTree()
    tree.parse(args.ofxfile)
    response = tree.convert()
    stmts = response.statements

    accounts = client.budget.be_accounts
    accountvsnotes={account.note:account for account in accounts if account.note is not None}

    for stmt in stmts:
        key = stmt.account.bankid + ' ' + stmt.account.branchid + ' ' + stmt.account.acctid
        if all(key not in note for note in accountvsnotes):
            if len(accounts) == 0:
                print('No accounts available in this budget')
                exit(-1)

            # ask user input for which bank account this is, then save it into the account note in nYNAB
            account = client.select_account_ui()

            # Save the selection in the nYNAB account note
            addon = 'key[' + key + ']key'
            if account.note is not None:
                account.note += addon
            else:
                account.note = addon
            client.budget.be_accounts.modify(account)
            client.sync()
        else:
            for note in accountvsnotes:
                if key in note:
                    account=accountvsnotes[note]

                    imported_date=datetime.now().date()

                    for ofx_transaction in stmt.transactions:
                        payee_name = ofx_transaction.name if ofx_transaction.payee is None else ofx_transaction.payee
                        try:
                            payee = next(p for p in client.budget.be_payees if p.name == payee_name)
                        except StopIteration:
                            payee=Payee(
                                name=payee_name
                            )
                            client.budget.be_payees.append(payee)
                            client.sync()

                        # use ftid so we don't import duplicates
                        if not any(ofx_transaction.fitid in transaction.memo for transaction in client.budget.be_transactions if
                                   transaction.memo is not None and transaction.entities_account_id == account.id):

                            transaction = Transaction(
                                date=ofx_transaction.dtposted,
                                memo=ofx_transaction.memo + '    '+ofx_transaction.fitid,
                                imported_payee=payee_name,
                                entities_payee_id=payee.id,
                                imported_date=imported_date,
                                source="Imported",
                                check_number=ofx_transaction.checknum,
                                amount=float(ofx_transaction.trnamt),
                                entities_account_id=account.id
                            )
                            client.add_transaction(transaction)

if __name__ == "__main__":
    ofximport_main()


import configargparse
import inspect
import os
import random
import re
import sys
from pynYNAB.config import test_common_args
from ynab import YNAB

from pynYNAB.Client import nYnabClient, BudgetNotFound, clientfromargs
from pynYNAB.Entity import AccountTypes
from pynYNAB.connection import nYnabConnection
from pynYNAB.schema.budget import MasterCategory, Subcategory, Account, Payee, Transaction


def migrate_main():
    print('migrate YNAB4 to pynYNAB')
    """Migrate a YNAB4 budget transaction history to nYNAB, using pyynab"""

    parser = configargparse.getArgumentParser('pynYNAB')
    parser.description=inspect.getdoc(migrate_main)
    parser.add_argument('budget', metavar='BudgetPath', type=str,
                        help='The budget .ynab4 directory')
    args = parser.parse_args()
    test_common_args(args)

    budget_base_name=os.path.basename(args.budget)
    budget_path=os.path.dirname(args.budget)
    budget_name=re.match(r"(?P<budget_name>.*)~[A-Z0-9]{8}\.ynab4",budget_base_name).groupdict().get('budget_name')

    if args.budgetname is not None:
        budget_name=args.budgetname

    thisynab = YNAB(budget_path,budget_name)

    client = clientfromargs(args, reset=True)



    for ynab4_account in thisynab.accounts:
        account=Account(
            name=ynab4_account.name,
            account_type=ynab4_account.type.value,
            on_budget=ynab4_account.on_budget,
            sortable_index=random.randint(-50000, 50000),
        )
        mindate=min([ynab4transaction.date for ynab4transaction in thisynab.transactions if ynab4transaction.account == ynab4_account])
        client.add_account(account, 0, mindate)

    for master_category in thisynab.master_categories:
        master_entity = MasterCategory(
            name=master_category.name,
            sortable_index=random.randint(-50000, 50000)
        )
        client.budget.be_master_categories.append(master_entity)
        for category in master_category.categories:

            entity = Subcategory(
                name=category.name,
                entities_master_category_id=master_entity.id,
                sortable_index=random.randint(-50000, 50000)
            )
            client.budget.be_subcategories.append(entity)
        client.sync()

    for ynab4_payee in thisynab.payees:
        payee=Payee(
            name=ynab4_payee.name
                   )
        client.budget.be_payees.append(payee)
        client.sync()

    for ynab4transaction in thisynab.transactions:
        transaction=Transaction(

        )
        pass
    #
    #     transactions = []
    #     subtransactions = []
    #     accumulatedSplits = []
    #     transfers = []
    #     reTransfer = re.compile('.*?Transfer : (?P<account>.*)')
    #     reSplit = re.compile('\(Split\ (?P<num1>\d+)\/(?P<num2>\d+)\)')
    #
    #     split_id = next(x.id for x in nYNABobject.budget.be_subcategories if x.internal_name == 'Category/__Split__')
    #
    #     for register_row in RegisterRows:
    #         try:
    #             payee_id = payeemapping[register_row.PayeeName]
    #         except KeyError:
    #             payee_id = None
    #         resultSplit = reSplit.match(register_row.Memo)
    #         if resultSplit:
    #             accumulatedSplits.append(register_row)
    #
    #             if resultSplit.group('num1') == resultSplit.group('num2'):
    #                 total = sum(map(lambda x: x.Inflow - x.Outflow, accumulatedSplits))
    #                 try:
    #                     payee_id = payeemapping[next(x.PayeeName for x in accumulatedSplits if x.PayeeName in payeemapping)]
    #                 except StopIteration:
    #                     payee_id = None
    #                 transaction = Transaction(
    #                     amount=total,
    #                     date=register_row.Date,
    #                     entities_payee_id=payee_id,
    #                     entities_account_id=accountmapping[register_row.AccountName],
    #                     entities_subcategory_id=split_id,
    #                     flag=register_row.Flag
    #                 )
    #                 transactions.append(transaction)
    #                 for split in accumulatedSplits:
    #                     result = reTransfer.match(split.PayeeName)
    #                     if result is not None:
    #                         otheraccount = result.group('account')
    #                         payee_id = payeemapping['Transfer : ' + otheraccount]
    #
    #                         # this => other
    #                         subtransaction = Subtransaction(
    #                             amount=split.Inflow - split.Outflow,
    #                             date=split.Date,
    #                             entities_payee_id=payee_id,
    #                             entities_account_id=accountmapping[split.AccountName],
    #                             transfer_account_id=accountmapping[otheraccount],
    #                             flag=split.Flag,
    #                             entities_transaction_id=transaction.id
    #                         )
    #                         subtransactions.append(subtransaction)
    #                         transfers.append(subtransaction)
    #                     else:
    #                         if split.MasterCategoryName == '':
    #                             # an out of budget transaction most probably
    #                             entities_subcategory_id = None
    #                         else:
    #                             entities_subcategory_id = submapping[split.MasterCategoryName][split.SubcategoryName]
    #                         subtransactions.append(Subtransaction(
    #                             amount=split.Inflow - split.Outflow,
    #                             date=split.Date,
    #                             entities_account_id=accountmapping[split.AccountName],
    #                             entities_payee_id=payee_id,
    #                             entities_subcategory_id=entities_subcategory_id,
    #                             flag=split.Flag,
    #                             entities_transaction_id=transaction.id
    #                         ))
    #                 accumulatedSplits = []
    #         else:
    #             result = reTransfer.match(register_row.PayeeName)
    #             if result is not None:
    #                 payee_id = payeemapping[register_row.PayeeName]
    #                 otheraccount = result.group('account')
    #                 # this => other
    #                 transaction = Transaction(
    #                     amount=register_row.Inflow - register_row.Outflow,
    #                     date=register_row.Date,
    #                     entities_payee_id=payee_id,
    #                     entities_account_id=accountmapping[register_row.AccountName],
    #                     transfer_account_id=accountmapping[otheraccount],
    #                     flag=register_row.Flag
    #                 )
    #                 transfers.append(transaction)
    #             else:
    #                 if register_row.MasterCategoryName == '':
    #                     # an out of budget transaction most probably
    #                     entities_subcategory_id = None
    #                 else:
    #                     entities_subcategory_id = submapping[register_row.MasterCategoryName][register_row.SubcategoryName]
    #                 transactions.append(Transaction(
    #                     amount=register_row.Inflow - register_row.Outflow,
    #                     date=register_row.Date,
    #                     entities_account_id=accountmapping[register_row.AccountName],
    #                     entities_payee_id=payee_id,
    #                     entities_subcategory_id=entities_subcategory_id,
    #                     flag=register_row.Flag
    #                 ))
    #
    #     unsplittransactions=[transaction for transaction in transactions if transaction.entities_subcategory_id != split_id]
    #     splittransactions=[transaction for transaction in transactions if transaction.entities_subcategory_id == split_id]
    #
    #     transactions_dict={transaction.id:transaction for transaction in transactions}
    #     subtransactions_dict={subtransaction.id:subtransaction for subtransaction in subtransactions}
    #
    #     transfers_dict = {tr.id: tr for tr in transfers}
    #     random.shuffle(transfers)
    #
    #     for i1 in range(len(transfers)):
    #         tr1 = transfers[i1]
    #         for i2 in range(len(transfers)):
    #             if i2 > i1:
    #                 tr2 = transfers[i2]
    #                 if isinstance(tr1, Transaction) and isinstance(tr2, Transaction):
    #                     if tr1.entities_account_id == tr2.transfer_account_id \
    #                             and tr1.date == tr2.date \
    #                             and tr1.amount == -tr2.amount:
    #
    #                         tr1.transfer_transaction_id = tr2.id
    #                         tr2.transfer_transaction_id = tr1.id
    #
    #                         transactions.append(tr1)
    #                         transactions.append(tr2)
    #                 elif isinstance(tr1, Transaction) and isinstance(tr2, Subtransaction):
    #                     tr2parent = next(x for x in transactions if x.id == tr2.entities_transaction_id)
    #                     if tr1.entities_account_id == tr2.transfer_account_id \
    #                             and tr1.date == tr2parent.date \
    #                             and tr1.amount == -tr2.amount:
    #                         siblings = [subtransaction for subtransaction in subtransactions if subtransaction.entities_transaction_id == tr2parent.id]
    #
    #                         tr1.transfer_subtransaction_id = tr2.id
    #                         tr2.transfer_transaction_id = tr1.id
    #
    #                         transactions.append(tr1)
    #                         subtransactions.append(tr2)
    #                 elif isinstance(tr1, Subtransaction) and isinstance(tr2, Transaction):
    #                     tr1parent = next(x for x in transactions if x.id == tr1.entities_transaction_id)
    #                     if tr1parent.entities_account_id == tr2.transfer_account_id \
    #                             and tr1parent.date == tr2.date \
    #                             and tr1.amount == -tr2.amount:
    #                         siblings = [subtransaction for subtransaction in subtransactions if subtransaction.entities_transaction_id == tr1parent.id]
    #
    #                         tr1.transfer_transaction_id = tr2.id
    #                         tr2.transfer_subtransaction_id = tr1.id
    #
    #                         subtransactions.append(tr1)
    #                         transactions.append(tr2)
    #
    #
    #
    #     nYNABobject.budget.be_transactions.extend(transactions)
    #     nYNABobject.budget.be_subtransactions.extend(subtransactions)
    #
    #     nYNABobject.sync()
    #
    #
    #
    #     pass

if __name__ == "__main__":
    migrate_main()